;; -*- lexical-binding: t; -*-

(require 'pdf-tools nil t)

(require 'jao-org-notes)
(require 'jao-doc-view)
(require 'jao-maildir)

(declare pdf-info-outline "pdf-info")

(defvar jao-org--sink-dir "./")
(defvar jao-org-open-pdf-fun #'jao-org--pdf-tools-open)

(defun jao-org--pdf-tools-open (path page &optional height)
  (org-open-file path 1)
  (jao-doc-view-goto-page page height))

(defun jao-org--pdf-open (path page &optional height)
  (funcall (or jao-org-open-pdf-fun #'jao-org--pdf-tools-open) path page height))

(defun jao-org-links--open-pdf (link)
  "Open LINK in pdf-view-mode."
  (cond ((string-match "\\(.*\\)::\\([0-9]*\\)\\+\\+\\([[0-9]\\.*[0-9]*\\)"  link)
         (let* ((path (match-string 1 link))
                (page (string-to-number (match-string 2 link)))
                (height (string-to-number (match-string 3 link))))
           (jao-org--pdf-open path page height)))
        ((string-match "\\(.*\\)::\\([0-9]+\\)$"  link)
         (let* ((path (match-string 1 link))
                (page (max 1 (string-to-number (match-string 2 link)))))
           (jao-org--pdf-open path page)))
        (t (jao-org--pdf-open link nil))))

(defun jao-org-links--follow-doc (link)
  (let* ((full-link (concat org-directory "/doc/" link))
         (dest-path (car (split-string full-link "::"))))
    (when (not (file-exists-p dest-path))
      (let* ((sink-file (expand-file-name link jao-org--sink-dir))
             (real-file (if (file-exists-p sink-file) sink-file
                          (read-file-name "Import file: "
                                          jao-org--sink-dir link link))))
        (rename-file real-file dest-path)))
    (if (jao-doc-view-is-pdf dest-path)
        (jao-org-links--open-pdf full-link)
      (browse-url (format "file://%s" (expand-file-name  dest-path))))))

(defun jao-org-links--complete-doc (&optional arg)
  (let ((default-directory jao-org--sink-dir))
    (let ((f (replace-regexp-in-string "^file:" "doc:"
                                       (org-file-complete-link arg))))
      (if (jao-doc-view-is-pdf f)
          (let ((page (read-from-minibuffer "Page: " "")))
            (if (> (string-to-number page) 0)
                (concat f "::" (read-from-minibuffer "Page: " ""))
              f))
        f))))

(defun jao-org-links-open-mail (fname)
  (let ((group (jao-maildir-file-to-group fname))
        (id (with-temp-buffer
              (insert-file fname)
              (goto-char (point-min))
              (message-mode)
              (not-modified)
              (message-fetch-field "Message-ID"))))
    (when (and group id) (org-gnus-follow-link group id))))

;;;###autoload
(defvar jao-org-links-pdf-store-fun nil)

(defun jao-org-links--store-pdf-link ()
  (or (when (fboundp jao-org-links-pdf-store-fun)
        (funcall jao-org-links-pdf-store-fun))
      (when (derived-mode-p 'pdf-view-mode 'doc-view-mode)
        (jao-org-links-store-pdf-link buffer-file-name
                                      (jao-doc-view-current-page)
                                      (jao-doc-view-section-title)))))

;;;###autoload
(defun jao-org-links-store-pdf-link (path page title)
  (org-store-link-props
   :type "doc"
   :link (format "doc:%s::%d" (file-name-nondirectory path) page)
   :description (format "%s (p. %d)" title page)))

;;;###autoload
(defun jao-org-insert-doc (title)
  (interactive "sDocument title: ")
  (insert (format "[[doc:%s][%s]]" (jao-doc-view-title->file title) title)))

;;;###autoload
(defun jao-org-org-to-pdf-file ()
  (expand-file-name (concat "doc/" (file-name-base buffer-file-name) ".pdf")
                    (file-name-directory jao-org-notes-dir)))

;;;###autoload
(defun jao-org-pdf-to-org-file (&optional file-name)
  (let* ((file-name (or file-name buffer-file-name))
         (bn (file-name-base file-name))
         (rx (format "%s\\.org$" (regexp-quote bn))))
    (save-some-buffers nil
                       (lambda ()
                         (string-prefix-p jao-org-notes-dir buffer-file-name)))
    (or (car (directory-files-recursively jao-org-notes-dir rx))
        (let* ((dirs (jao-org-notes-cats))
               (dir (completing-read "Notes subdir: " dirs nil t)))
          (expand-file-name (concat dir "/" bn ".org") jao-org-notes-dir)))))

;;;###autoload
(defun jao-org-insert-doc-skeleton (&optional title)
  (insert "#+title: " (or title (jao-doc-view-title (buffer-file-name)))
          "\n#+author:\n#+filetags: ")
  (jao-org-notes-insert-tags)
  (insert  "\n#+startup: latexpreview\n\n"))

;;;###autoload
(defun jao-org-pdf-goto-org (arg)
  (interactive "P")
  (when (jao-doc-view-is-pdf buffer-file-name)
    (let* ((file (jao-org-pdf-to-org-file))
           (new (not (file-exists-p file)))
           (title (jao-doc-view-title)))
      (when (or arg new) (org-store-link nil t))
      (find-file-other-window file)
      (when new
        (jao-org-insert-doc-skeleton title)
        (org-insert-link)))))

;;;###autoload
(defun jao-org-pdf-goto-org* () (interactive) (jao-org-pdf-goto-org t))

;;;###autoload
(defun jao-org-org-goto-pdf ()
  (interactive)
  (if-let (f (jao-org-org-to-pdf-file))
      (find-file-other-window f)
    (user-error "No PDF file associated with this buffer")))

;;;###autoload
(defun jao-org-links-setup (sink-dir)
  (interactive)
  (org-link-set-parameters "doc"
                           :follow #'jao-org-links--follow-doc
                           :complete #'jao-org-links--complete-doc
                           :store #'jao-org-links--store-pdf-link)
  (org-link-set-parameters "docview" :store #'ignore)
  (org-link-set-parameters "message" :follow #'jao-org-links-open-mail)
  (setq jao-org--sink-dir (file-name-as-directory sink-dir)))

(provide 'jao-org-links)
